#%%
#!/bin/python3
#%%
import json
import os
import time
import pandas as pd
import numpy as np
import hashlib

from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split
from xgboost import XGBClassifier

from tools.modeling.evaluation_tools import series_auc_and_ks_score
from tools.std_lib.log import ColorfulLogger

from tools.std_lib.basic import json_load, json_dump, get_uuid_hex
from tools.data_processing.basic_data_tools import silence_pickle_load, silence_pickle_dump, dict_to_plain
from tools.data_processing.easy_pandas import get_hash, get_timestamp, t_print

import warnings
from pandas.api.types import is_sparse

warnings.filterwarnings("ignore", category=FutureWarning, module="sklearn.utils.validation")

#%%
log = ColorfulLogger(basename='training_tools_v2.py')

#%%

def get_qcut_points(s: pd.Series, n = 20):
    s = s.sort_values().dropna()
    cut_points = [s.iloc[int(len(s)*(k/n))] for k in range(1, n)]
    cut_points = sorted(set(cut_points))
    return cut_points


def count_qcut(s: pd.Series, n = 20):
    cut_points = get_qcut_points(s, n)
    return len(cut_points) + 1

#%%



def easy_make_classification(
    n_sample = 1000,
    n_features = 10,
    n_informative = None,
    n_classes = 2,
    class_sep = .5,

    n_redundant = 0,
    flip_y = 0,
    random_state = 0,
    n_clusters_per_class = 1,
    weights = [.8, .2],
    test_size = .4
    ):

    # informative, redundant and repeated features

    if n_informative is None:
        n_informative = n_features

    X, y = make_classification(
        n_samples=n_sample,
        n_features=n_features,
        n_informative=n_informative,
        n_redundant = n_redundant,
        n_classes=n_classes,
        class_sep=class_sep,
        flip_y=flip_y,
        random_state=random_state,
        n_clusters_per_class=n_clusters_per_class,
        weights=weights,
    )

    idx = pd.Index(range(1, n_sample+1), name='index')

    X_df = pd.DataFrame(X)
    X_df.columns = [f'x{k}' for k in range(n_features)]
    X_df.index = idx

    y_df = pd.DataFrame(y)
    y_df.columns = ['y']
    y_df.index = idx

    print(f'n_sample = {n_sample}')
    print(f'n_features = {n_features}')
    print(f'weights = {weights}')
    print('y_pos_pct =', y_df['y'].mean())

    X_train_df, X_test_df, y_train_s, y_test_s = train_test_split(
        X_df, 
        y_df['y'], 
        test_size=test_size, 
        random_state=0, 
        stratify=y_df['y'],
    )
    train_idx = y_train_s.index
    validate_idx = y_test_s.index[:int(len(y_test_s) // 2)  ]
    test_idx     = y_test_s.index[ int(len(y_test_s) // 2): ]
    y_df['data_split'] = ''
    y_df.loc[train_idx, 'data_split'] = 'train'
    y_df.loc[validate_idx, 'data_split'] = 'validate'
    y_df.loc[test_idx, 'data_split'] = 'test'
    print('value_counts.value_counts:', y_df['data_split'].value_counts())

    return X_df, y_df



#%%



def find_dir_with_prefix(prefix: str, dir_path: str):
    dst_paths = []
    for name in os.listdir(dir_path):
        dst_path = os.path.join(dir_path, name)
        if name.startswith(prefix) and os.path.isdir(dst_path):
            dst_paths.append(dst_path)
    if len(dst_paths) == 0:
        result = None
    elif len(dst_paths) == 1:
        result = dst_paths[0]
    else:
        raise ValueError(f"Multiple directories found with prefix {prefix} in {dir_path}: {dst_paths}")
    return result


def get_classifier_from_name(classifier_name: str, classifier_params = {}):
    func_name = 'get_classifier_from_name()'
    if isinstance(classifier_name, str):
        if classifier_name in ['LR']:
            log.info(f'classifier = LogisticRegression(**{classifier_params})', color = 'green', func = func_name)
            classifier = LogisticRegression(**classifier_params)
        elif classifier_name in ['XGB']:
            log.info(f'classifier = XGBClassifier(**{classifier_params})', color = 'green', func = func_name)
            classifier = XGBClassifier(**classifier_params)
        else:
            raise ValueError(f'Unknown classifier name {classifier_name}')
    else:
        raise TypeError(f'Expected str for classifier_name, got {type(classifier_name)}')
    return classifier


class BinaryClassifierTrainer(object): 
    '''

    '''
    def __init__(self, 
                 X_train_df: pd.DataFrame,
                 y_train_s: pd.Series,
                 classifier_name: str,
                 classifier_params = {},
                 prefix  = None,
                 cache_dir = None,
                 ):
        self.X_train_df = X_train_df
        self.y_train_s = y_train_s
        self.classifier_name = classifier_name
        self.clf = get_classifier_from_name(classifier_name, classifier_params)
        self.prefix = prefix
        self.cache_dir = cache_dir
        return None
    

    def get_train_hash(self):
        train_info = {
            'X': {
                'index': self.X_train_df.index.tolist(),
                'columns': self.X_train_df.columns.tolist()
            },
            'y': {
                'index': self.y_train_s.index.tolist(),
                'values': self.y_train_s.tolist()
            },
            'classifier_name': self.classifier_name,
            'all_classifier_params': dict_to_plain(self.clf.get_params())
        }
        train_hash = get_hash(json.dumps(train_info, sort_keys=True))
        return train_hash
    
    def get_train_exp_name(self):
        train_hash = self.get_train_hash()
        exp_name = f'{self.classifier_name}_{train_hash[:16]}'
        if self.prefix is not None:
            exp_name = f'{self.prefix}-{exp_name}'
        log.info(f'train_exp_name = \'{exp_name}\'', func = 'trainer.get_train_exp_name()')
        return exp_name
    
    def check_and_fix(self):
        func_name = 'trainer.check_and_fix()'
        if not self.X_train_df.index.equals(self.y_train_s.index):
            raise ValueError('X_train_df.index != y_train_s.index')
        if not self.X_train_df.index.is_unique:
            raise ValueError('X_train_df.index is not unique')
        if not self.X_train_df.columns.is_unique:
            raise ValueError('X_train_df.columns is not unique')
        if not self.y_train_s.index.equals(self.X_train_df.index):
            raise ValueError('y_train_s.index != X_train_df.index')
        if not self.y_train_s.index.is_unique:
            raise ValueError('y_train_s.index is not unique')
        
        if 'random_state' in dir(self.clf):
            if self.clf.random_state is None:
                self.clf.random_state = 0
                log.info('set clf.random_state = 0', func = func_name)
        log.info('OK', func = func_name)
        self.exp_name = self.get_train_exp_name()
        return None
    
    def exists(self):
        flag = False
        cache_dir = self.cache_dir
        if cache_dir is not None and isinstance(cache_dir, str):
            if self.prefix is None:
                prefix_path = cache_dir
            else:
                prefix_path = os.path.join(cache_dir, self.prefix)
            if os.path.exists(prefix_path) and os.path.isdir(prefix_path):
                exp_name = self.exp_name
                save_path = find_dir_with_prefix(exp_name, prefix_path)
                if save_path is not None:
                    self.save_path = save_path
                    flag = True
        log.info(f'flag = {flag}', func = 'trainer.exists()')
        return flag
    
    def train(self):
        func_name = 'trainer.train()'
        log.info(f'start at T = 0s', func = func_name)
        t0 = time.time()
        self.clf.fit(self.X_train_df.to_numpy(), self.y_train_s.to_numpy())
        t1 = time.time()
        delta_time = t1 - t0
        log.info(f'completed at T = {delta_time:.3f}s', func = func_name)
        return None

    def dump(self):
        cache_dir = self.cache_dir
        if cache_dir is not None :
            exp_name = self.exp_name
            timestamp = get_timestamp()
            save_name = f'{exp_name}-{timestamp}'
            prefix = self.prefix
            if prefix is None:
                save_path = os.path.join(cache_dir, save_name)
            else:
                save_path = os.path.join(cache_dir, prefix, save_name)
            clf_pickle_path = os.path.join(save_path, 'clf.pickle')
            os.makedirs(save_path, exist_ok=True)
            silence_pickle_dump(self.clf, clf_pickle_path)
            self.save_path = save_path
            log.info(f'dumped, save_path = \'{save_path}\'', func = 'trainer.dump()')
        return None

    def load(self):
        clf_pickle_path = os.path.join(self.save_path, 'clf.pickle')
        self.clf = None
        self.clf = silence_pickle_load(clf_pickle_path)
        log.info(f'loaded, save_path = \'{self.save_path}\'', func = 'trainer.load()')
        return None
        
    def run(self):
        func_name = 'trainer.run()'
        log.info('start', func = func_name)
        self.check_and_fix()
        if not self.exists():
            self.train()
            self.dump()
        else:
            self.load()
        log.info('completed', func = func_name)
        return None

    pass # end


def auc_ks_groupby(proba_df: pd.DataFrame, by: str, sorted_index = None)->pd.DataFrame:
    gp = proba_df.groupby(by = by)
    stat_df = gp.apply(lambda df: series_auc_and_ks_score(df['y'], df['proba']))
    stat_df.sort_index(inplace = True)
    if sorted_index is not None:
        stat_df = stat_df.loc[sorted_index, :].copy()
    return stat_df


def get_qcut_points(s: pd.Series, n = 20):
    s = s.sort_values().dropna()
    cut_points = [s.iloc[int(len(s)*(k/n))] for k in range(1, n)]
    cut_points = sorted(set(cut_points))
    return cut_points


def count_qcut(s: pd.Series, n = 20):
    cut_points = get_qcut_points(s, n)
    return len(cut_points) + 1


def get_lift_from_proba_df(proba_df: pd.DataFrame, top: float):
    top_value = proba_df['proba'].sort_values().iloc[-int(len(proba_df)*top)]
    top_proba_df = proba_df.loc[proba_df['proba'] >= top_value, :]
    lift = top_proba_df['y'].mean() / proba_df['y'].mean()
    return lift


class BinaryClassifier(object):
    '''
    prefix: (defined by user)
    uuid_hex: (auto)
    exp_name: prefix-XGB_1234567890abcdef
    save_name: prefix-XGB_1234567890abcdef-20250413Sun0101
    predict_proba_name: (=prefix, default)

    exp_name: prefix-X[S01_maas_r1+S02+S03+S04_bp]-XGB_12345678
    save_name: prefix-X[S01_maas_r1+S02+S03+S04_bp]-XGB_12345678-20250413Sun0101

    '''
    def __init__(self,
                 X_df: pd.DataFrame,
                 y_df: pd.DataFrame,
                 features : list = None,
                 fillna: object = None,
                 classifier_name = 'XGB',
                 classifier_params = {},
                 prefix = None,
                 cache_dir = None,
                 tag = None,
                 train_cache_dir = None,
                 ):
        self.X_df = X_df
        self.y_df = y_df.copy()
        self.features = features
        self.fillna = fillna
        self.classifier_name = classifier_name
        self.classifier_params = classifier_params
        self.classifier_params = self.fix_classifier_params()
        self.prefix = prefix
        self.cache_dir = cache_dir
        self.tag = tag
        self.clf = None
        self.train_cache_dir = train_cache_dir
        return None
    

    def fix_classifier_params(self):
        y_df = self.y_df
        classifier_params = self.classifier_params.copy()
        if self.classifier_name in ['XGB']:
            train_y_s = y_df.loc[y_df['data_split'] == 'train', 'y']
            scale_pos_weight = (train_y_s == 0).sum() / (train_y_s == 1).sum()
            classifier_params['scale_pos_weight'] = scale_pos_weight
        elif self.classifier_name in ['LR']:
            classifier_params['class_weight'] = 'balanced'
        return classifier_params

    def X_df_fillna(self):
        if self.classifier_name in ['LR']:
            if self.fillna is None:
                train_idx = self.y_df['data_split'] == 'train'
                self.fillna = self.X_df.loc[train_idx, :].mean().to_dict()
            else:
                raise ValueError('fillna must be None when classifier_name = LR.')
            self.X_df = self.X_df.fillna(self.fillna)
            if self.X_df.isna().any().any():
                raise ValueError('fillna failed when LR.')
        else:
            if self.fillna is None:
                self.fillna = {k: float('nan') for k in self.X_df.columns}
            elif isinstance(self.fillna, str):
                if self.fillna == 'train_mean':
                    train_idx = self.y_df['data_split'] == 'train'
                    self.fillna = self.X_df.loc[train_idx, :].mean().to_dict()
                    self.X_df = self.X_df.fillna(self.fillna)
                    if self.X_df.isna().any().any():
                        raise ValueError('fillna failed when not LR.')
                else:
                    raise ValueError(f'self.fillna = {self.fillna} is not supported.')
            else:
                raise ValueError(f'self.fillna = {self.fillna} is not supported.')
        return None
    

    def get_clf_hash(self):
        train_info = {
            'X': {
                'index': self.X_df.index.tolist(),
                'columns': self.X_df.columns.tolist()
            },
            'y': self.y_df.to_dict(),
            'classifier_name': self.classifier_name,
            'all_classifier_params': dict_to_plain(self.classifier_params),
            'fillna': self.fillna,
            'tag': self.tag,
        }
        train_hash = get_hash(json.dumps(train_info, sort_keys=True))
        return train_hash
    
    def get_exp_name(self):
        train_hash = self.get_clf_hash()
        exp_name = f'{self.classifier_name}_{train_hash[:16]}'
        if self.prefix is not None:
            exp_name = f'{self.prefix}-{exp_name}'
        log.info(f'exp_name = \'{exp_name}\'', func = 'clf.get_exp_name()')
        return exp_name
    

    def get_split_items(self):
        old_splits = self.y_df['data_split'].dropna().unique().tolist()
        main_splits = ['train', 'validate', 'test']
        new_splits = [item for item in main_splits if item in old_splits]
        new_splits += sorted([item for item in old_splits if item not in main_splits])
        return new_splits


    def check_and_fix(self):
        if not self.X_df.index.equals(self.y_df.index):
            raise ValueError("X_df and y_df must have the same index")

        if self.y_df['y'].dtype.name != 'int32':
            raise ValueError(f'y.dtype.name is not int32')

        if sorted(self.y_df['y'].unique().tolist()) != [0, 1]:
            raise ValueError("y_df must have at least two unique values [0, 1]")
        if 'month' not in self.y_df.columns:
            log.warning(f'no month in y_df and will set to 1970-01')
            self.y_df['month'] = '1970-01'
        
        if self.y_df.isna().any().any():
            raise ValueError("y_df must not contain any NaN values")
        
        if self.features is None:
            self.features = self.X_df.columns.to_list()
        self.features = sorted(self.features)
        self.X_df = self.X_df[self.features]
        
        self.X_df_fillna()
        self.exp_name = self.get_exp_name()
        self.BinaryClassifierTrainer = BinaryClassifierTrainer
        self.splits = self.get_split_items()

        if 'train' not in self.splits:
            raise ValueError(f'data_split must contain "train"')

        if not self.X_df.columns.is_monotonic_increasing:
            raise ValueError(f'X_df.columns.is_monotonic_increasing = False')

        return None
    
    
    def set_trainer(self, trainer: object):
        self.BinaryClassifierTrainer = trainer
        return None
    
    def set_evaluator(self, evaluator: object):
        self.evaluate = evaluator
        pass
 
    def exists(self):
        flag = False
        cache_dir = self.cache_dir
        if cache_dir is not None and isinstance(cache_dir, str):
            if self.prefix is None:
                prefix_path = cache_dir
            else:
                prefix_path = os.path.join(cache_dir, self.prefix)
            if os.path.exists(prefix_path) and os.path.isdir(prefix_path):
                exp_name = self.exp_name
                save_path = find_dir_with_prefix(exp_name, prefix_path)
                if save_path is not None:
                    self.save_path = save_path
                    flag = True
        log.info(f'flag = {flag}', func = 'clf.exists()')
        return flag
    
    def train(self):
        X_df = self.X_df
        y_df = self.y_df
        train_idx = y_df['data_split'] == 'train'
        X_train_df = X_df.loc[train_idx, :]
        y_train_s = y_df.loc[train_idx, 'y']
        trainer = self.BinaryClassifierTrainer(
            X_train_df = X_train_df,
            y_train_s = y_train_s,
            classifier_name = self.classifier_name,
            classifier_params = self.classifier_params,
            prefix  = self.prefix,
            cache_dir = self.train_cache_dir,
        )
        trainer.run()
        self.clf = trainer.clf
        return None

    def predict(self):
        X_np = self.X_df.to_numpy()
        proba_df = self.y_df.copy()
        proba_df['proba'] = self.clf.predict_proba(X_np)[:, 1]
        self.proba_df = proba_df
        return None

    def evaluate(self):
        proba_df = self.proba_df
        splits = self.splits
        auc_ks_by_data_split = auc_ks_groupby(
            proba_df=proba_df, 
            by='data_split', 
            sorted_index=splits
            )
        auc_ks_by_month = auc_ks_groupby(
            proba_df=proba_df, 
            by='month', 
            )
        
        y_cnt_stat_df = proba_df[['data_split', 'y']].value_counts().sort_index().unstack(level = 'y')
        y_cnt_stat_df.columns = [f'y={k}' for k in y_cnt_stat_df.columns]
        y_cnt_stat_df['count'] = y_cnt_stat_df.sum(axis=1)
        y_cnt_stat_df['pos_pct'] = y_cnt_stat_df['y=1'] / y_cnt_stat_df['count']
        y_cnt_stat_df = y_cnt_stat_df.loc[splits]

        bins_cnt_by_data_split = proba_df.groupby(by = 'data_split')['proba'].apply(
            lambda s: pd.Series({
                'bins_cnt_10' :count_qcut(s, 10),
                'bins_cnt_20' :count_qcut(s, 20),
                'bins_cnt_50' :count_qcut(s, 50),
                })
            ).unstack(level=1).loc[splits]
        
        simple_describe_by_data_split = proba_df.groupby('data_split')['proba'].describe().loc[splits]
        simple_describe_by_data_split['count'] = simple_describe_by_data_split['count'].astype(int)

        lift_by_data_split = proba_df.groupby('data_split').apply(lambda df: pd.Series({
            f'lift_{k:02d}': get_lift_from_proba_df(df, top = k / 100)
                for k in [
                    1,  2,  3,  4,  5, 
                    10, 20, 30, 40, 50,
            ]})).loc[splits]
        
        evaluation_dict = {}
        evaluation_dict['auc_ks_by_data_split'] = auc_ks_by_data_split
        evaluation_dict['auc_ks_by_month'] = auc_ks_by_month
        evaluation_dict['lift_by_data_split'] = lift_by_data_split
        evaluation_dict['y_cnt_by_data_split'] = y_cnt_stat_df
        evaluation_dict['bins_cnt_by_data_split'] = bins_cnt_by_data_split
        evaluation_dict['simple_describe_by_data_split'] = simple_describe_by_data_split

        self.evaluation_dict = evaluation_dict

        return None
    
    def show(self):
        evaluation_dict = self.evaluation_dict
        log.info(
            'auc_ks_by_data_split:', 
            evaluation_dict['auc_ks_by_data_split'].round(3), 
            color = 'green'
            )
        log.info(
            'auc_ks_by_month:', 
            evaluation_dict['auc_ks_by_month'].round(3), 
            color = 'green'
            )
        log.info(
            'lift_by_data_split:', 
            evaluation_dict['lift_by_data_split'].round(3), 
            color = 'green'
            )
        log.info(
            'bins_cnt_by_data_split:', 
            evaluation_dict['bins_cnt_by_data_split'].round(3), 
            color = 'green'
            )
        log.info(
            'simple_describe_by_data_split:', 
            evaluation_dict['simple_describe_by_data_split'].round(3), 
            color = 'green'
            )

    
    def get_save_path(self):
        cache_dir = self.cache_dir
        if cache_dir is not None :
            exp_name = self.exp_name
            timestamp = get_timestamp()
            save_name = f'{exp_name}-{timestamp}'

            self.save_name = save_name
            self.timestamp = timestamp
            self.uuid_hex = get_uuid_hex()

            prefix = self.prefix
            if prefix is None:
                save_path = os.path.join(cache_dir, save_name)
            else:
                save_path = os.path.join(cache_dir, prefix, save_name)
        else:
            save_path = None

        return save_path
    

    def get_xgb_feature_importances(self):
        feature_importances_dict = dict(zip(
            self.features, 
            self.clf.feature_importances_
            ))
        feature_importances_dict = dict_to_plain(feature_importances_dict)
        return feature_importances_dict
    
    def get_clf_info_json(self):

        d = {}

        d['main'] = {}
        d['main']['classifier_name'] = self.classifier_name
        d['main']['prefix'] = self.prefix
        d['main']['exp_name'] = self.exp_name
        d['main']['save_name'] = self.save_name
        d['main']['timestamp'] = self.timestamp
        d['main']['uuid_hex'] = self.uuid_hex
        d['main']['X_df.shape'] = list(self.X_df.shape)
        d['main']['y_df.shape'] = list(self.y_df.shape)

        d['clf'] = {}
        d['clf']['my_params'] = self.classifier_params
        d['clf']['clf_str'] = self.clf.__str__().replace('\n', '').replace(' ', '')
        d['clf']['params'] = self.clf.get_params()

        d['feature'] = {}
        d['feature']['features'] = self.features
        d['feature']['fillna'] = self.fillna

        if self.classifier_name in ['XGB']:
            d['feature']['feature_importances_info'] = {}
            feature_importances_dict = self.get_xgb_feature_importances()
            d['feature']['feature_importances_info']['feature_importances'] = feature_importances_dict
            d['feature']['feature_importances_info']['n_features_in_use'] = len({
                k: v for k, v in feature_importances_dict.items() if v > 0
            })

        evaluation_dict = self.evaluation_dict
        auc_ks_by_data_split = evaluation_dict['auc_ks_by_data_split']
        auc_ks_by_month = evaluation_dict['auc_ks_by_month']

        auc_ks_by_split_dict = auc_ks_by_data_split.T.unstack().to_dict()
        auc_ks_by_split_dict = {'_'.join(k): v for k, v in auc_ks_by_split_dict.items()}
        
        auc_ks_by_month_dict = auc_ks_by_month.T.unstack().to_dict()
        auc_ks_by_month_dict = {'_'.join(k): v for k, v in auc_ks_by_month_dict.items()}
        
        d['performance'] = {}
        d['performance']['auc_and_ks_by_data_split'] = auc_ks_by_split_dict
        d['performance']['auc_and_ks_by_month'] = auc_ks_by_month_dict

        lift_by_data_split = self.evaluation_dict['lift_by_data_split']
        lift_by_data_split_dict = lift_by_data_split.T.unstack().to_dict()
        lift_by_data_split_dict = {'-'.join(k): v for k, v in lift_by_data_split_dict.items()}
        d['performance']['lift_by_data_split'] = lift_by_data_split_dict

        bins_cnt_by_data_split_dict = evaluation_dict['bins_cnt_by_data_split'].T.unstack().to_dict()
        bins_cnt_by_data_split_dict = {'-'.join(k): v for k, v in bins_cnt_by_data_split_dict.items()}
        d['performance']['bins_cnt_by_data_split'] = bins_cnt_by_data_split_dict

        self.info_json = dict_to_plain(d)

        return d
        
        
    def dump(self):
        '''
        '''
        save_path = self.get_save_path()

        if save_path is not None :

            os.makedirs(save_path, exist_ok=True)

            clf_pickle_path = os.path.join(save_path, 'clf.pickle')
            silence_pickle_dump(self.clf, clf_pickle_path)
            
            features_json_path = os.path.join(save_path, 'features.json')
            json_dump(self.features, features_json_path)

            fillna_json_path = os.path.join(save_path, 'fillna.json')
            json_dump(self.fillna, fillna_json_path)

            evaluation_pickle_path = os.path.join(save_path, 'evaluation.pickle')
            silence_pickle_dump(self.evaluation_dict, evaluation_pickle_path)

            info_json_path = os.path.join(save_path, 'info.json')
            self.get_clf_info_json()
            json_dump(self.info_json, info_json_path)

            self.save_path = save_path
            log.info(f'dumped, save_path = \'{save_path}\'', func = 'clf.dump()')

        return None

    def load(self):
        '''
        '''
        save_path = self.save_path

        clf_pickle_path = os.path.join(save_path, 'clf.pickle')
        self.clf = silence_pickle_load(clf_pickle_path)

        features_json_path = os.path.join(save_path, 'features.json')
        self.features = json_load(features_json_path)

        fillna_json_path = os.path.join(save_path, 'fillna.json')
        self.fillna = json_load(fillna_json_path)

        evaluation_pickle_path = os.path.join(save_path, 'evaluation.pickle')
        self.evaluation_dict = silence_pickle_load(evaluation_pickle_path)

        info_json_path = os.path.join(save_path, 'info.json')
        self.info_json = json_load(info_json_path)
        
        log.info(f'loaded, save_path = \'{self.save_path}\'', func = 'clf.load()')
        return None

    def run(self):
        self.check_and_fix()
        if not self.exists():
            self.train()
            self.predict()
            self.evaluate()
            self.dump()
        else:
            self.load()
        self.show()
        return None

    pass # end




#%%
# import base64
# import datetime
# import functools
# import gzip
# import json
# import os
# from itertools import product
# from random import shuffle
# import re
# from time import time, sleep

# import numpy as np
# import pandas as pd

# from sklearn.linear_model import LogisticRegression
# from sklearn.metrics import log_loss
# from sklearn.model_selection import train_test_split
# from sklearn.ensemble import GradientBoostingClassifier
# from sklearn.ensemble import HistGradientBoostingClassifier
# from xgboost import XGBClassifier
# from lightgbm import LGBMClassifier

# # from flaml import AutoML
# from hyperopt import hp, fmin, tpe, Trials, STATUS_OK, space_eval

# from sklearn2pmml import sklearn2pmml
# from sklearn2pmml.pipeline import PMMLPipeline
# from pypmml import Model

# from tools.data_processing.basic_data_tools import pickle_load
# from tools.std_lib.basic import json_dumps, json_dump, json_load
# from tools.data_processing.easy_pandas import dict_to_json, get_bytes_sha256, get_file_hash, get_timestamp, ls_dirs, ls_dirs_r, multiprocessing_apply, pickle_dump, read_txt, save_and_reset_index, t_print, get_hash, date_str, get_uuid_hex, save_txt

##%%
# def check_xy_data(X = [], y = [], found_nan = 'raise'):
#     for Xk in X:
#         print('Xmin =', Xk.min().min())
#         print('Xmax =', Xk.max().max())
#         if Xk.isna().any().any() and found_nan == 'raise':
#             raise ValueError('X has nan')
#         elif Xk.isna().any().any():
#             print('X has nan')
#         else:
#             print('X has no nan')
#     for yk in y:
#         print(yk.value_counts().to_dict())
#         if yk.isna().any() :
#             raise ValueError('y has nan')
#     return None

# def split_xy_data(y_df: pd.DataFrame, x_df: pd.DataFrame, how = 'intersec', fillna = -10, found_nan='raise'):
    
#     t_print('fillna =', fillna)
#     y_df = y_df.loc[~y_df['data_split'].isna(), :]
#     idx = y_df.index.intersection(x_df.index)
#     n1 = len(y_df)
#     n2 = len(idx)
#     t_print(f'coverage: {n2 / n1} (= {n2}/{n1})')
#     y_df = y_df.loc[idx, :]
#     y_df['y'] = y_df['y'].astype(int)
#     x_df = x_df.loc[idx, :].astype(float)
    
#     if isinstance(fillna, str) and fillna == 'mean':
#         fillna = x_df.loc[y_df['data_split'] == 'train', :].mean()
#         t_print('fillna =', fillna.to_dict())
#     if fillna is not None:
#         x_df = x_df.fillna(fillna)
    
#     split_ls = ['train', 'validate', 'test']
#     y_dfs = [y_df.loc[y_df['data_split'] == split, 'y'] for split in split_ls]
#     x_dfs = [x_df.loc[y_df['data_split'] == split, :  ] for split in split_ls]
    
#     t_print(f'y_dfs = {y_dfs}', )
#     t_print(f'x_dfs = {x_dfs}', )
    
#     check_xy_data(x_dfs, y_dfs, found_nan=found_nan)
    
#     return y_dfs, x_dfs

# def get_xy_tri(X, y, ds = 0, test_size = .2):
#     if len(X) == len(y) == 2:
#         X_sample, X_test = X
#         y_sample, y_test = y
#         X_train, X_valid, y_train, y_valid = train_test_split(X_sample, y_sample, test_size=test_size, random_state=ds, stratify=y_sample)
#     elif len(X) == len(y) == 3:
#         X_train, X_valid, X_test = X
#         y_train, y_valid, y_test = y
#         if ds != 0:
#             del test_size
#             test_size = len(y_valid) / (len(y_train) + len(y_valid))
#             t_print('(estimated) test_size =', test_size, ', random_stat =', ds)
#             X_sample = pd.concat([X_train, X_valid]).sort_index()
#             y_sample = pd.concat([y_train, y_valid]).sort_index()
#             X_train, X_valid, y_train, y_valid = train_test_split(X_sample, y_sample, test_size=test_size, random_state=ds, stratify=y_sample)
#     elif len(X) == len(y) > 3:
#         X_sample, X_test, y_sample, y_test = train_test_split(X, y, test_size=test_size, random_state=0, stratify=y)
#         X_train, X_valid, y_train, y_valid = train_test_split(X_sample, y_sample, test_size=test_size, random_state=ds, stratify=y_sample)
#     return X_train, X_valid, X_test, y_train, y_valid, y_test


# #%%
# class AutoTrainMany(object):


#     def __init__(self, 
#                  X_df: pd.DataFrame,
#                  y_df: pd.DataFrame, # with data_split
#                  allinone_params: list = None,
#                  test_only = False,
#                  add_pmml = True,
#                  fillna = None,
#                  ):
#         self.X = X_df
#         self.y = y_df
#         self.allinone = allinone_params
#         self.test_only = test_only
#         self.fillna = fillna
#         self.add_pmml = add_pmml
    
#     def split_to_xy_data(self):
#         if isinstance(self.X,pd.DataFrame):
#             if 'data_split' in self.y:
#                 y, X = split_xy_data(y_df=self.y, x_df=self.X, fillna=self.fillna, found_nan='ignore')
#                 self.X,self.y = X,y
#         elif isinstance(self.X,tuple):
#             self.X = list(self.X)
#             self.y = list(self.y)
#         elif isinstance(self.X,list):
#             ds = self.allinone.get('random_state',0)
#             test_size = self.allinone.get('test_size',.2)
#             X_train, X_valid, X_test, y_train, y_valid, y_test = get_xy_tri(self.X,self.y,ds,test_size)
#             self.X = [X_train,X_valid,X_test]
#             self.y = [y_train,y_valid,y_test]




#     def get_params_dict_list(self):
#         all_keys = list(self.allinone.keys())
#         for k in all_keys:
#             if self.allinone[k] == []:
#                 del self.allinone[k]

#         keys = self.allinone.keys()
#         values = [k if type(k) == list else [k] for k in self.allinone.values()]
#         params_dict_list = [dict(zip(keys, combination)) for combination in product(*values)]
        
#         r = []
#         for k in params_dict_list:
#             if k not in r:
#                 if k['clf_tag'] == 'AutoML':
#                     if k['AutoML.metric_alpha'] * k['AutoML.cv_score_agg_func_alpha'] != 0:
#                         continue
#                 r.append(k)

#         return shuffle(r)
    
#     def clean_xy(self):

#         X,y,fillna = self.X,self.y,self.fillna

#         X = [Xk.copy().astype(float) for Xk in X]
#         y = [yk.copy() for yk in y]
        
#         if not (len(X) == len(y) == 3):
#             raise ValueError('X and y must be list of length 3')

#         if isinstance(fillna, int) or isinstance(fillna, float) :
#             fillna = {k: fillna for k in X[0].keys()}

#         if isinstance(fillna, str) and fillna == 'mean':
#             fillna = X[0].mean()

#         if isinstance(fillna, pd.Series):
#             fillna = fillna.to_dict()
#             print('fillna =', fillna)

#         t_print('fillna')
#         [Xk.fillna(fillna, inplace = True) for Xk in X]

#         if not isinstance(fillna, dict):
#             raise ValueError('fillna must be dict or pd.Series')
        
#         self.X_df,self.y_df,self.fillna = X,y,fillna

#     def Train_many(self):
#         self.clean_xy()
#         eta,t0,settings_dict_list = 0,time(),self.get_params_dict_list()

#         global exp_name_global
#         global iter_global

#         for k, settings in enumerate(settings_dict_list):
#             t1 = time()
            
#         t_print(f'All will end at {date_str(eta)}.\n#')

#         pass


# class AutoTrainOnce(object):


#     def __init__(self, 
#                  X_df: pd.DataFrame,
#                  y_df: pd.DataFrame, # with data_split, date 
#                  saving_params: dict = None, # prefix and results_path
#                  data_split_params: dict = None,
#                  training_params: dict = None,
#                  evaluation_params: dict = None,
#                  ):
#         self.X_df = X_df
#         self.y_df = y_df
#         self.saving_params = saving_params
#         self.test_date = data_split_params.get('test_date', None)
#         self.data_split_seed = data_split_params.get('data_split_seed', None)
#         self.training_params = training_params
#         self.evaluation_params = evaluation_params
#         pass

